/*
 * File:   main.c
 * Author: Lucas Volwater 
 *
 * Created on May 15, 2023, 3:34 PM
 */

/* PWR usage (re-test when changing fw)
 * all at 3.7V
 * time sync - about 5 mA
 * time display - about 25 - 30  mA
 * idle / waiting for gesture to display: about 0.5 mA
 * Battery is 110 mAh, so will last about a week / 9 days with no display so a week with a few displays
 */

#include <xc.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include "rtc.h"
#include <util/crc16.h> // using crc8, contained therein.
#include <avr/sleep.h>
#include <avr/power.h>
#define F_CPU 8e6 // for timing/delays, even though it is running at 1 MHz part of the time

#define LDTHRESH 200 // 0-255 light/dark threshold for setting time /data reception. Adapt for screen brightness.
// 200 for my screen.

#define DIFFTHRESH 5 // 0-255 threshold fir light/dark difference for gesture detecion for 'display switch on'. 
// 20 seems to work but  not in dim environment. 
// 5 works in dim and bright environment but might cause false triggers. todo: Long term test?


FUSES = {
    .low = 0xE2, //E2 = INTRC 8MHZ/1 = 8MHZ. (default = 0x62=INTRC 8 MHZ/8 =1Mhz.) 3V is enough for 8 MHz (2.7V for up to 10, even) 
    .high = 0xD9, // D9=default. to enable debugwire use 0x99
    .extended = 0xFF, // geen BOD
};

LOCKBITS = 0xFF; // {LB=PROG_VER_DISABLED, BLB0=LPM_SPM_DISABLE, BLB1=LPM_SPM_DISABLE}

uint8_t read_time_optical();
void display_hours(uint8_t);
void display_min_sec(uint8_t);
void display_binairy(uint8_t);
/*either light sensor OR smps feedback can be sampled by ADC, also a few other settings need to be made for this*/
void disable_smps_and_enable_lightsensor();
void disable_lightsensor_and_start_smps();
void waitabit(); // waits 1 tick of TCNT2, the RTC timer ticking at 256 Hz. So waits about 4ms.
void battcheck();

#define DEBUGLED_ON do{PORTD |= 1 << PORTD7;}while(0);
#define DEBUGLED_OFF do{PORTD &= ~(1 << PORTD7);}while(0);

volatile uint8_t adcresult, recd_data, count_light, count_dark, count_bits; /* used in adc interrupt but declared global for easy access */
volatile int16_t diff;

int main(void) {
    uint8_t h, m, s;
    /*init*/
    /*IO for Atmega328
     * PC0=ADC0=SMPS feedback
     * PC1=ADC1=Light sensor
     * ADC7 is battery voltage after a 220k/33k divider with its GND on PORTD.1
     * PB6/PB7 = XTAL 32kHz
     * PD5=OC0B=SMPS PWM
     * PC2 Light sensor pwr
     * Neons on the folowing pins: 
     * PB0, PB3, PB4, PB5
     * PC4,PC5,
     * PD0,PD2
     * PORTD.7 is a debug LED
     * PORTD.1 is divider GND for battery voltage
     */

    clock_prescale_set(clock_div_1); // start at 8Mhz

    DDRB = (1 << DDB0) | (1 << DDB3) | (1 << DDB4) | (1 << DDB5); // outputs for neons
    DDRC = (1 << DDC2) | (1 << DDC4) | (1 << DDC5); // Light sensor power, neon outputs
    DDRD = (1 << DDD5) | (1 << DDD0) | (1 << DDD2) | (1 << DDD7) | (1 << DDD1); // PWM output, neon outputs, divider gnd for battery

    /*power consumption*/
    DIDR0 = (1 << ADC0D) | (1 << ADC1D); // disable digital input buffer on ADC pins to minimize power consumption
    PORTD |= (1 << PORTD1); // divider disable

    /*timers*/
    /*RTC timer 2, clock 32.768kHz xtal*/
    ASSR = (1 << AS2); // first set clock to async operation, as that might corrupt registers, and since these are set later it does not matter
    TCCR2A = 0x00; // default / normal mode / no output
    TCCR2B = (1 << CS22) | (1 << CS20); // divide clock by 128, so 32768/128/256 overflow each second.
    TIFR2 = 0x07; // clear interrupt flags
    TIMSK2 = (1 << TOIE2); // enable overflow interrupt

    sei(); //enable interrupts
    
    battcheck(); // check battery on boot first.

    disable_lightsensor_and_start_smps();
    DEBUGLED_ON;
    //show each bit, so it is clear if bit location is correct
    display_binairy(1);
    display_binairy(2);
    display_binairy(4);
    display_binairy(8);
    display_binairy(16);
    display_binairy(32);
    display_binairy(64);
    display_binairy(128);
    disable_smps_and_enable_lightsensor();
    DEBUGLED_OFF;
    

#if 1 // disable to debug power usage and wake-gesture
    while (!read_time_optical()) {
    } // sync time on reboot
#endif
    /*loop*/

    while (1) {

        static enum {
            STAGE1, STAGE2, STAGE3, DETECTED
        } detect = DETECTED; //start at detected, so time gets displayed directly after setting it.
        static uint8_t timeout, now, batt_check_time;
        /*
         * detect dark/light/dark sequence as wakeup to turn the display on,
         * reduce clock to save power in the meantime (maybe could sleep instead?)
         */

        now = TCNT2; // to get somewhat of a fixed sample rate, use Timer 2 (RTC timer, increments at 256 Hz = overflows each second).        
        ADCSRA |= (1 << ADSC); // start a adc conversion for light sensor. 
        timeout++;
        batt_check_time++;
        switch (detect) { // detect change in light level / hand wave above watch, and then display the time.
            /* higher lightsensor reading is darker. So positive diff means it got brighter and neg. diff means it got darker */
            //TODO: test this: maybe besides difference in light level, also use 'dark' (cover watch with hand) to reduce false triggers
          
            case STAGE1:
                timeout = 0;
                if (diff< -DIFFTHRESH) detect = STAGE2;
                break;
            case STAGE2:
                if (diff > DIFFTHRESH) detect = STAGE3;
                break;
            case STAGE3:
                if ((diff< -DIFFTHRESH) && (adcresult >= LDTHRESH)) detect = DETECTED;
                break;
            case DETECTED:
                //enable SMPS and display time
                disable_lightsensor_and_start_smps();

                for (uint8_t i = 0; i < 28; i++) {
                    gettime_h(&h);
                    display_hours(h);
                }
                for (uint8_t i = 0; i < 2; i++) {
                    display_binairy(0);
                }
                for (uint8_t i = 0; i < 28; i++) {
                    gettime_m(&m);
                    display_min_sec(m);
                }
#if 0 // displaying seconds is more distracting then usefull
                for (uint8_t i = 0; i < 2; i++) {
                    display_binairy(0);
                }
                for (uint8_t i = 0; i < 25; i++) {
                    gettime_s(&s);
                    display_min_sec(s);
                }
#endif
                disable_smps_and_enable_lightsensor();
                detect = STAGE1;
            default:
                detect = STAGE1;
        }
        clock_prescale_set(clock_div_128); // reduce clock further here, to reduce power consumption
        //now at 0.5 mA, depending on Vcc (@3.7V))
        while ((uint8_t) (TCNT2 - now) < 32) {
            // wait 32/256 = 1/8 of a second, so 8 Samples/s. (125 ms)
        };
        clock_prescale_set(clock_div_8); // restore to 1 MHz
        if (timeout > 24) { // 24* 1/8s = 24/8s=3s
            detect = STAGE1;
        }
        if (batt_check_time >= 160) { // Every 160 loops == every 20 seconds, check te battery
            batt_check_time = 0;
            battcheck();
            disable_smps_and_enable_lightsensor(); // to switch ADC back to lightsensor ch.
        }
    }
    return 0;
}

uint8_t read_time_optical() {
    static uint8_t receive_buffer[4], now, succes, crc = 0;

    enum {
        READH, READM, READS, READCRC, CHECKCRC
    } state = READH;
    succes = 0;
    disable_smps_and_enable_lightsensor();
    do {
        now = TCNT2; // to get somewhat of a fixed sample rate, use Timer 2 (RTC timer, increments at 256 Hz = overflows each second).
        ADCSRA |= (1 << ADSC); // start a adc conversion.
        while ((uint8_t) (TCNT2 - now) < 4) {
            // wait 4/256 = 1/64 of a second, so 64 Samples/s. (15.625 ms)
        };
        if (adcresult > LDTHRESH) {  // debug led to aid in determening display threshold calibration and where to hold the watch during sync
            DEBUGLED_ON;
        } else DEBUGLED_OFF;
        switch (state) {
            default:
                state = READH;
                crc = 0; //reset crc
            case READH:
                if (count_bits == 8) {
                    receive_buffer[0] = recd_data;
                    state = READM;
                }
                break;
            case READM:
                if (count_bits == 16) {
                    receive_buffer[1] = recd_data;
                    state = READS;
                }
                break;
            case READS:
                if (count_bits == 24) {
                    receive_buffer[2] = recd_data;
                    state = READCRC;
                }
                break;
            case READCRC:
                if (count_bits == 32) {
                    receive_buffer[3] = recd_data;
                    state = CHECKCRC;
                }
                break;
            case CHECKCRC:
                //crc is calculated here as not to spend time on it during reception.
                crc = _crc8_ccitt_update(crc, receive_buffer[0]);
                crc = _crc8_ccitt_update(crc, receive_buffer[1]);
                crc = _crc8_ccitt_update(crc, receive_buffer[2]);

                if (crc == receive_buffer[3]) {// if crc matches, set time to new received value, else retry      
                    settime_hms(receive_buffer[0], receive_buffer[1], receive_buffer[2]);
                    succes = 1;
                } else {
                    crc = 0;
                    count_bits = 0;
                    state = READH; // wait for syncword again after time is set or also on CRC mismatch
                }
                break;
        }
        // once CRC is OK and time is set, return.
    }while (!succes);
    return succes;
}

void disable_smps_and_enable_lightsensor() {
    /*disable PWM output from timer0. Maybe even switch back to 1 MHz*/
    PORTD &= ~(1 << PORTD5); // set IO to 0.
    TCCR0A = 0x00; // set to IO instead of PWM output
    /* lower clock for lower power consumption. */
    clock_prescale_set(clock_div_8); // set to /8, so 1MHz

    ADCSRA = 0; // disable/reset ADC

    /*set adc to correct ch and samplerate for autotrigger (adc clock), or disable auto trigger*/
    ADMUX = (1 << ADLAR) | (1 << REFS0) | (1 << MUX0); // vcc as adc reference, ADC ch1, left adjust so only upper 8 bits hace to be read
    ADCSRA = (1 << ADIE) | (1 << ADSC) | (1 << ADEN) | (1 << ADPS2); // start ADC, enable interrupt, single conversion, clock /16 (so 1M/16=62.5 kHz);
    //ADCSRA = (1 << ADIE) | (1 << ADSC) | (1 << ADEN) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // start ADC, enable interrupt, single conversion, clock /128 (so 8M/128=62.5 kHz);

    PORTC |= (1 << PORTC2); /*Turn sensor on*/

}

void disable_lightsensor_and_start_smps() {
    PORTC &= ~(1 << PORTC2); /* turn sensor off */
    /*switch to 8MHz*/
    clock_prescale_set(clock_div_1);

    ADCSRA = 0; // disable/reset ADC

    /*set adc to correct ch and samplerate for autotrigger (adc clock)*/
    ADMUX = (1 << ADLAR) | (1 << REFS1) | (1 << REFS0); // ch0, left adjust (8bit),1.1V Vref
    ADCSRA = 0xFF; // enable ADC, start first conversion, auto trigger enabled (free running), CLKIO/128 as clock (8M/128=62.5 kHz), interrupt enabled (and flag cleared)

    /*setup PWM timer (OC0B) and output
     * Clock from CLK1O, no prescaler.
     * TOP = OCR0A = 80. So on 8MHz PWM freq will be 8M/80= 100 kHz and OCR1B ranges 80-0 (inverting PWM)
     */
    OCR0A = 80;
    OCR0B = 80; // start at 0% ducy
    TCNT0 = 0;
    TCCR0A = (1 << COM0B0) | (1 << COM0B1) | (1 << WGM00) | (1 << WGM01); // OC0B inverting, fast PWM mode 2 (OCR1A=TOP)
    TCCR0B = (1 << WGM02) | (1 << CS00);

    //If PWM frequency is measured to be 12.5 kHz, chip is running at 1 Mhz instead of 8. Symptoms include a failing SMPS.
    // Somehow changing CKDIV8 and CLKPR has no effect, and CLKPR stays 3 in variable watch window.
    // Restarting the debug session on another chip shows no such issues, and after trying that it also works on the original chip again.
    // So it is a 'weird issue'- probably power supply or breadboard - next time try just poweroff/poweron, re-seating the chip and unpluggin/re-plugging the debugger.

}

void display_hours(uint8_t h) {
    //display hours (folows analogue clock, see note/sletch)
    if (h > 12) h = h - 12;
    switch (h) {
        default:
            //this should not happen
            display_binairy(0xff); // all on on error
            break;
        case 1:
            display_binairy(0x02);
            break;
        case 2:
            display_binairy(0x04);
            break;
        case 3:
            display_binairy(0x0C);
            break;
        case 4:
            display_binairy(0x08);
            break;
        case 5:
            display_binairy(0x10);
            break;
        case 6:
            display_binairy(0x30);
            break;
        case 7:
            display_binairy(0x20);
            break;
        case 8:
            display_binairy(0x40);
            break;
        case 9:
            display_binairy(0xC0);
            break;
        case 10:
            display_binairy(0x80);
            break;
        case 11:
            display_binairy(0x01);
            break;
        case 12:
        case 0: // 0 and 12 are equivalent
            display_binairy(0x03);
            break;
    }
}

void display_min_sec(uint8_t m) {
    // Display minutes or seconds (per 5, see note/sketch)
    m = (m + 2) / 5; // simple rounding to 'nearest' 5. (hint: it is an uint)
    switch (m) {
        case 0: // 0-2 minutes (or seconds)
            display_binairy(0x03);
            break;
        case 1: // 3-7
            display_binairy(0x02);
            break;
        case 2: // 8-12
            display_binairy(0x06);
            break;
        case 3: // 13-17
            display_binairy(0x0C);
            break;
        case 4: // 18-22
            display_binairy(0x08);
            break;
        case 5: // 23-27
            display_binairy(0x18);
            break;
        case 6: // 28-32
            display_binairy(0x30);
            break;
        case 7: // 33-37
            display_binairy(0x20);
            break;
        case 8: // 38-42
            display_binairy(0x60);
            break;
        case 9: // 43-47
            display_binairy(0xC0);
            break;
        case 10: // 48-52
            display_binairy(0x80);
            break;
        case 11: // 53-57
            display_binairy(0x81);
            break;
        case 12: // 57-00 (62)
            display_binairy(0x87); // just before the next hour, to avoid confusion with start of this hour, make it something special
            break;
        default:
            display_binairy(0xFF); //error
            break;
    }
}

void display_binairy(uint8_t b) {
    // raw binairy display (bit 0 is top left and then numbered clockwise)
    // 1 neon indicator electrode at a time
    for (uint8_t i = 0; i < 8; i++) {
        if (b & 0x01) PORTB |= (1 << PORTB0); //PORTB0=b.0, Neon 0,0
        waitabit();
        PORTB &= ~(1 << PORTB0);
        if (b & 0x02) PORTB |= (1 << PORTB3); //PORTB3=b.1, Neon 0,1
        waitabit();
        PORTB &= ~(1 << PORTB3);
        if (b & 0x04) PORTB |= (1 << PORTB5); //portb5=b.2, Neon 1,0
        waitabit();
        PORTB &= ~(1 << PORTB5);
        if (b & 0x08) PORTB |= (1 << PORTB4); //portb4=b.3, Neon 1,1
        waitabit();
        PORTB &= ~(1 << PORTB4);
        if (b & 0x10) PORTC |= (1 << PORTC4); //portc4=b.4, Neon 2,0
        waitabit();
        PORTC &= ~(1 << PORTC4);
        if (b & 0x20) PORTC |= (1 << PORTC5); //portc5=b.5, Neon 2,1
        waitabit();
        PORTC &= ~(1 << PORTC5);
        if (b & 0x40) PORTD |= (1 << PORTD0); //portd0=b.6, Neon 3,0
        waitabit();
        PORTD &= ~(1 << PORTD0);
        if (b & 0x80) PORTD |= (1 << PORTD2); //portd2=b.7, Neon 3,1
        waitabit();
        PORTD &= ~(1 << PORTD2);
    }
}

void waitabit() {
    uint8_t now = TCNT2;
    while (now == TCNT2); // wait 1 counter tick. (32768Hz/128) = 256 Hz, so 3.9 ms, plus a bit for call overhead
}

void battcheck() {
    /* with 1v1 reference and a 220k/33k divider
     * 3V * 33k/253k = 0.39
     * 256*0.39/1.1=90
     * so if adcresult<90, power down to save the battery (force a recharge before the protection circuit kicks in)
     */
    PORTD &= ~(1 << PORTD1); // divider enable
    while (ADCSRA & _BV(ADSC)) {
        //iff there currently is an adc conversion going on, finish that before changing settings
    };
    ADMUX = (1 << ADLAR) | (1 << REFS0) | (1 << REFS1) | (1 << MUX0) | (1 << MUX1) | (1 << MUX2); // 1v1 adc reference, ADC ch7, left adjust so only upper 8 bits hace to be read
    ADCSRA = (1 << ADSC) | (1 << ADEN) | (1 << ADPS2) | (1 << ADIF); // start ADC, single conversion, clock /16 (so 1M/16=62.5 kHz), clear flag

    //wait untill conversion is done
    while ((ADCSRA & (1 << ADSC))) {
    };

    /* 
     * datasheet: The first ADC conversion result after switching reference voltage source
     * may be inaccurate, and the user is advised to discard this result
     * 
     * Reality: AREF needs to fall from VCC to 1v1, this takes time too.
     * 
     */

    // waitabit is about 4 ms, so this is about 12:
    waitabit();
    waitabit();
    waitabit();

    ADCSRA |= (1 << ADSC); // then start 2nd conversion
    while ((ADCSRA & (1 << ADSC))) {
        //wait untill adc conversion is done
    };

    PORTD |= (1 << PORTD1); // divider disable  
#if 0
    if (ADCH < 91) {
        DEBUGLED_ON;
    } else DEBUGLED_OFF;
#endif

#if 1
    if (ADCH < 91) { //if adcresult<3V then ZZZzzz
        //set all I/O to the most low-power state posible, go to deepest sleep posible
        cli();
        DEBUGLED_OFF;
        PORTD &= ~(1 << PORTD5); // set IO to 0.
        TCCR0A = 0x00; // set to IO instead of PWM output (Stop SMPS)
        PORTC = 0; // Turn sensor off
        PORTD |= (1 << PORTD1); // voltage divider OFF
        sleep_bod_disable();
        ADCSRA = 0; //disable ADC
        PRR = 0xEF; // stop all modules
        set_sleep_mode(SLEEP_MODE_PWR_DOWN);
        sleep_mode(); // <0.02 mA
#endif
    }
}

ISR(TIMER2_OVF_vect) {
    /*RTC timing with Timer2. Triggers at 1 Hz.*/
    clocktick(); // one second has passed
}

ISR(ADC_vect) {
    /* ADC for light sensor and voltage feedback.
     * 
     * for voltage feedback there is a 1M/2k2  1:455 divider and a 1.1V reference voltage, and a 8 bit result.
     * So 1.1V/256 = 4,3 mV per tick * 455 = 1,96 V per tick -> 75 ticks is about 147 V.
     * there is quite a voltage swing when neons load the smps, so keep enough margin to ignite the neon
     *
     */

    diff = adcresult; // use previous result to calculate difference
    adcresult = ADCH; //read adc
    diff -= adcresult;

    /* check which channel it was to decide what to do with it */
    if ((ADMUX & 0x0F) == 0x01) { //if adc_channel == 1 == lightsensor
        /*
         * used for light/dark detection (display on) and to sync time using PWM
         * '1' is 40 ms black folowed by 80 ms white (1-4 samples of bk, 5-7 samples of white)
         * '0' is 80 ms black folwed by 40 ms white (5-7 samples of bk, 1-4 samples of white)
         * reset to bit 0 on longer then 156 ms white (sampletime is 15.6 ms, so, 10 samples)
         */

        if (adcresult > LDTHRESH) { // thresholding.
            count_dark++;
            count_light = 0;
        } else {
            //on reception of first light, check
            if (count_dark > 1) {
                recd_data = recd_data >> 1; // always shift the buffer.
                count_bits++; // count number of bits shifted in
                if (count_dark < 5) { // 1,2,3,4 samples dark it is a '1'. (5,6,7,+ a '0').
                    recd_data = recd_data | 0x80;
                }
            }
            count_dark = 0; //reset counter
            count_light++; // sample count is used for timing.
            if (count_light >= 10) { // when line is idle for too long, reset
                count_light = 0;
                count_bits = 0;
            }
        }

    } else if ((ADMUX & 0x0F) == 0x00) { //if adc channel == 0 == smps feedback
        if (adcresult < 75) {
            if (OCR0B > (80 - 12)) OCR0B -= 1; // limit to 15% dutycycle max. (12 is 15% of 80, 80 is max count, but PWM is inverted so 80-12). 
            // rise dutycycle until error is 0.
        } else {
            if (OCR0B < 80) OCR0B += 1;
        }
    }
}